@@ -1,6 +1,5 @@
 import math
 import threading
-from tempfile import TemporaryDirectory
 
 import numpy as np
 from PyQt5.QtGui import QBrush, QPen, QColor, QPalette, QPainter, QPainterPath
@@ -17,10 +16,10 @@ from circlevis.beatmap_info import BeatmapInfo
 from circlevis.player import Player
 
 WIDTH_LINE = 1
+WIDTH_LINE_RAW_VIEW = 2
 WIDTH_CROSS = 2
-WIDTH_CIRCLE_BORDER = 8
+WIDTH_CIRCLE_BORDER = 6
 LENGTH_CROSS = 6
-FRAMES_ON_SCREEN = 15  # how many frames for each replay to draw on screen at a time
 
 PEN_WHITE = QPen(QColor(200, 200, 200))
 PEN_GRAY = QPen(QColor(75, 75, 75))
@@ -33,8 +32,8 @@ BRUSH_GRAY = QBrush(QColor(100, 100, 100))
 BRUSH_DARKGRAY = QBrush(QColor(10, 10, 10))
 BRUSH_BLANK = QBrush(QColor(0, 0, 0, 0))
 
-GAMEPLAY_PADDING_WIDTH = 64 + 192
-GAMEPLAY_PADDING_HEIGHT = 48 + 48
+GAMEPLAY_PADDING_WIDTH = 64 + 60
+GAMEPLAY_PADDING_HEIGHT = 48 + 20
 GAMEPLAY_WIDTH = 512
 GAMEPLAY_HEIGHT = 384
 
@@ -45,15 +44,19 @@ SLIDER_TICKRATE = 50
 
 class Renderer(QFrame):
     update_time_signal = pyqtSignal(int)
+    pause_signal = pyqtSignal()
     analyzer = RunTimeAnalyser(frame_buffer=FRAMETIME_FRAMES)
 
-    def __init__(self, beatmap_info, replays, events, library, start_speed, paint_info):
+    def __init__(self, beatmap, replays, events, library, start_speed, \
+        paint_info, statistic_functions):
         super().__init__()
         self.setMinimumSize(GAMEPLAY_WIDTH + GAMEPLAY_PADDING_WIDTH*2, GAMEPLAY_HEIGHT + GAMEPLAY_PADDING_HEIGHT*2)
         # list of timestamps to highlight the frames of in a different color
         self.events = events
         # whether to show some information about each player and their cursors
         self.should_paint_info = paint_info
+        # functions to display info for in the visualizer
+        self.statistic_functions = statistic_functions
         # whether we should paint the frametime graph
         self.paint_frametime = False
         self.painter = QPainter()
@@ -62,44 +65,26 @@ class Renderer(QFrame):
         self.y_offset = 0
 
         # beatmap init stuff
-        self.hitobjs = []
+        self.hitobjs_to_draw = []
 
-        if beatmap_info.path:
-            self.beatmap = Beatmap.from_path(beatmap_info.path)
-            self.hit_objects = self.beatmap.hit_objects()
+        if beatmap:
+            self.hit_objects = beatmap.hit_objects()
             self.playback_len = self.get_hit_endtime(self.hit_objects[-1])
-        elif beatmap_info.map_id:
-            # library is nullable - None means we define our own (and don't care about saving)
-            # TODO move temporary directory creation to slider probably, since
-            # this logic is now duplicated here and in circlecore
-            if library:
-                # TODO expose save as an option to the user somehow?
-                # might require a slider pr or just a change in approach for us
-                self.beatmap = library.lookup_by_id(beatmap_info.map_id, download=True, save=True)
+            # for now we'll use the hr or dt modified cs, ar, and od if any
+            # replay has HR enabled
+            use_hr = any([Mod.HR in replay.mods for replay in replays])
+            ar = beatmap.ar(hard_rock=use_hr)
+            # https://osu.ppy.sh/help/wiki/Beatmapping/Approach_rate for formulas
+            if ar <= 5:
+                self.preempt = 1200 + 600 * (5 - ar) / 5
+                self.fade_in = 800 + 400 * (5 - ar) / 5
             else:
-                temp_dir = TemporaryDirectory()
-                self.beatmap = Library(temp_dir.name).lookup_by_id(beatmap_info.map_id, download=True)
-            self.hit_objects = self.beatmap.hit_objects()
-            self.playback_len = self.get_hit_endtime(self.hit_objects[-1])
-        else:
-            self.playback_len = 0
+                self.preempt = 1200 - 750 * (ar - 5) / 5
+                self.fade_in = 800 - 500 * (ar - 5) / 5
 
-        self.has_beatmap = beatmap_info.available()
+            self.hitwindow = od_to_ms(beatmap.od(hard_rock=use_hr)).hit_50
 
-
-        # beatmap stuff
-        if self.has_beatmap:
-            # values taken from https://github.com/ppy/osu-wiki/blob/master/meta/unused/difficulty-settings.md
-            # but it was taken from the osu! wiki since then so this might be a bit incorrect.
-            if self.beatmap.approach_rate == 5:
-                self.preempt = 1200
-            elif self.beatmap.approach_rate < 5:
-                self.preempt = 1200 + 600 * (5 - self.beatmap.approach_rate) / 5
-            else:
-                self.preempt = 1200 - 750 * (self.beatmap.approach_rate - 5) / 5
-            self.hitwindow = od_to_ms(self.beatmap.overall_difficulty).hit_50
-            self.fade_in = 400
-            self.hitcircle_radius = circle_radius(self.beatmap.circle_size) - WIDTH_CIRCLE_BORDER / 2
+            self.hitcircle_radius = circle_radius(beatmap.cs(hard_rock=use_hr))
             ## loading stuff
             self.is_loading = True
             # not fully accurate, but good enough
@@ -108,9 +93,11 @@ class Renderer(QFrame):
             self.sliders_current = 0
             self.thread = threading.Thread(target=self.process_sliders)
             self.thread.start()
-
+            self.has_beatmap = True
         else:
+            self.playback_len = 0
             self.is_loading = False
+            self.has_beatmap = False
 
         # if this is nonnull, when we finish loading sliders we will seek to
         # this position. Set in ``seek_to`` if it is called when we're loading
@@ -142,7 +129,6 @@ class Renderer(QFrame):
         self.timer = QTimer(self)
         self.timer.timeout.connect(self.next_frame_from_timer)
         self.timer.start(1000/60) # 62 fps (1000ms/60frames but the result can only be a integer)
-        self.next_frame()
 
         # black background
         pal = QPalette()
@@ -150,6 +136,18 @@ class Renderer(QFrame):
         self.setAutoFillBackground(True)
         self.setPalette(pal)
 
+        # Settings that are changeable from the control's setting button.
+        # If `True`, don't draw crosses, and draw the line in grey if the user
+        # was not pressing any keys in the start frame of that line.
+        self.raw_view = False
+        self.draw_hitobjects = True
+        self.draw_approach_circles = True
+        # how many frames for each replay to draw on screen at a time
+        self.num_frames_on_screen = 15
+        self.only_embolden_keydowns = False
+
+        self.next_frame()
+
     def resizeEvent(self, event):
         width = event.size().width() - GAMEPLAY_PADDING_WIDTH * 2
         height = event.size().height() - GAMEPLAY_PADDING_HEIGHT * 2
@@ -212,25 +210,27 @@ class Renderer(QFrame):
         # if we're at the end of the track or are at the beginning of the track
         # (and thus are reversing), pause and dont update
         if current_time > self.playback_len or current_time < 0:
-            self.pause()
+            self.pause_signal.emit()
             return
 
         for player in self.players:
             player.end_pos = np.searchsorted(player.t, current_time, "right") - 1
-            player.start_pos = player.end_pos - FRAMES_ON_SCREEN if player.end_pos >= FRAMES_ON_SCREEN else 0
+            player.start_pos = player.end_pos - self.num_frames_on_screen if player.end_pos >= self.num_frames_on_screen else 0
 
         if self.has_beatmap:
             self.get_hitobjects()
         self.update_time_signal.emit(current_time)
         self.update()
 
-    @analyzer.track
     def get_hitobjects(self):
         # get current hitobjects
         current_time = self.clock.get_time()
         found_all = False
+        # TODO optimize this by tracking our current hitobj index, this iterates
+        # through half the hitobjects of the map on average (O(1) best case and
+        # O(n) worst case) which can't be good for performance
         index = 0
-        self.hitobjs = []
+        self.hitobjs_to_draw = []
         while not found_all:
             current_hitobj = self.hit_objects[index]
             hit_t = current_hitobj.time.total_seconds() * 1000
@@ -239,7 +239,7 @@ class Renderer(QFrame):
             else:
                 hit_end = hit_t + self.hitwindow + self.fade_in
             if hit_t - self.preempt < current_time < hit_end:
-                self.hitobjs.append(current_hitobj)
+                self.hitobjs_to_draw.append(current_hitobj)
             elif hit_t > current_time:
                 found_all = True
             if index == self.num_hitobjects - 1:
@@ -282,7 +282,6 @@ class Renderer(QFrame):
         self.analyzer.enabled = self.paint_frametime
         self.painter.end()
 
-    @analyzer.track
     def paint_cursor(self, player):
         """
         Draws a cursor.
@@ -290,10 +289,11 @@ class Renderer(QFrame):
         Arguments:
             Player player: player to draw the cursor of.
         """
-        alpha_step = 1 / FRAMES_ON_SCREEN
+        alpha_step = 1 / self.num_frames_on_screen
         pen = player.pen
-        pen.setWidth(self.scaled_number(WIDTH_LINE))
-        PEN_HIGHLIGHT.setWidth(self.scaled_number(WIDTH_LINE))
+        width = WIDTH_LINE_RAW_VIEW if self.raw_view else WIDTH_LINE
+        pen.setWidth(self.scaled_number(width))
+        PEN_HIGHLIGHT.setWidth(self.scaled_number(width))
         self.painter.setPen(pen)
         highlighted_pen = False
         for i in range(player.start_pos, player.end_pos):
@@ -304,8 +304,19 @@ class Renderer(QFrame):
             elif not highlight and highlighted_pen:
                 self.painter.setPen(pen)
                 highlighted_pen = False
-            self.draw_line((i-player.start_pos) * alpha_step, (player.xy[i][0], player.xy[i][1]),
-                           (player.xy[i + 1][0], player.xy[i + 1][1]))
+            grey_out = False
+            # only grey out lines if we're in raw view (crosses are greyed out
+            # instead in the normal view)
+            if self.raw_view:
+                # grey out if we don't have a keypress at the start
+                if not bool(player.k[i]):
+                    grey_out = True
+                # grey out if we're only emboldening keydowns and this is not a
+                # keydown
+                if self.only_embolden_keydowns and not bool(player.keydowns[i]):
+                    grey_out = True
+            self.draw_line((i - player.start_pos) * alpha_step, player.xy[i],
+                    player.xy[i + 1], grey_out=grey_out)
         pen.setWidth(self.scaled_number(WIDTH_CROSS))
         self.painter.setPen(pen)
         for i in range(player.start_pos, player.end_pos+1):
@@ -314,15 +325,20 @@ class Renderer(QFrame):
             k = player.k[i]
             t = player.t[i]
             highlight = t in self.events
-            self.draw_cross(alpha, xy, grey_out = not bool(k), highlight=highlight)
+            # grey out only if no keys are held by default
+            grey_out = not bool(k)
+            # but override if we're only emboldening keydowns and this is not a
+            # keydown
+            if self.only_embolden_keydowns and not bool(player.keydowns[i]):
+                grey_out = True
+            self.draw_cross(alpha, xy, grey_out=grey_out, highlight=highlight)
         # reset alpha
         self.painter.setOpacity(1)
 
     def paint_beatmap(self):
-        for hitobj in self.hitobjs[::-1]:
+        for hitobj in self.hitobjs_to_draw[::-1]:
             self.draw_hitobject(hitobj)
 
-    @analyzer.track
     def paint_info(self):
         """
         Draws various Information.
@@ -330,6 +346,10 @@ class Renderer(QFrame):
         Args:
            QPainter painter: The painter.
         """
+        # our current y coordinate for drawing info. Modified throughout this
+        # function
+        y = 15
+
         PEN_WHITE.setWidth(self.scaled_number(1))
         self.painter.setPen(PEN_WHITE)
         self.painter.setOpacity(0.25)
@@ -337,30 +357,41 @@ class Renderer(QFrame):
         PEN_WHITE.setWidth(1)
         self.painter.setPen(PEN_WHITE)
         self.painter.setOpacity(1)
-        self.painter.drawText(5, 15, f"Clock: {round(self.clock.get_time())} ms | Cursor count: {len(self.players)}")
+        self.painter.drawText(5, y, f"{round(self.clock.get_time())} ms")
+
         if self.num_replays > 0:
-            for i, player in enumerate(self.players):
+            for player in self.players:
+                y += 13
                 pen = player.pen
                 self.painter.setPen(PEN_BLANK)
                 self.painter.setBrush(QBrush(pen.color()))
                 self.painter.setOpacity(1 if Key.M1 in Key(int(player.k[player.end_pos])) else 0.3)
-                self.painter.drawRect(5, 27 - 9 + (11 * i), 10, 10)
+                self.painter.drawRect(5, y - 9, 10, 10)
                 self.painter.setOpacity(1 if Key.M2 in Key(int(player.k[player.end_pos])) else 0.3)
-                self.painter.drawRect(18, 27 - 9 + (11 * i), 10, 10)
+                self.painter.drawRect(18, y - 9, 10, 10)
                 self.painter.setOpacity(1)
                 self.painter.setPen(pen)
-                self.painter.drawText(31, 27 + (11 * i), f"{player.username} {player.mods.short_name()}: {player.xy[player.end_pos][0]:.2f}, {player.xy[player.end_pos][1]:.2f}")
+                self.painter.drawText(31, y, f"{player.username} {player.mods.short_name()}: {player.xy[player.end_pos][0]:.2f}, {player.xy[player.end_pos][1]:.2f}")
+
             self.painter.setPen(PEN_WHITE)
             if self.num_replays == 2:
                 try:
-                    player = self.players[1]
-                    prev_player = self.players[0]
-                    distance = math.sqrt(((prev_player.xy[prev_player.end_pos][0] - player.xy[player.end_pos][0]) ** 2) +
-                                         ((prev_player.xy[prev_player.end_pos][1] - player.xy[player.end_pos][1]) ** 2))
-                    self.painter.drawText(5, 39 + (12 * 1), f"Distance {prev_player.username}-{player.username}: {int(distance)}px")
+                    y += 13
+                    p1 = self.players[0]
+                    p2 = self.players[1]
+                    distance = math.sqrt(((p1.xy[p1.end_pos][0] - p2.xy[p2.end_pos][0]) ** 2) +
+                                         ((p1.xy[p1.end_pos][1] - p2.xy[p2.end_pos][1]) ** 2))
+                    self.painter.drawText(5, y, f"{int(distance)}px apart")
                 except IndexError: # Edge case where we only have data from one cursor
                     pass
 
+            for function in self.statistic_functions:
+                y += 13
+                xys = [player.xy for player in self.players]
+                indices = [player.end_pos for player in self.players]
+                result = function(xys, indices)
+                self.painter.drawText(5, y, f"{function.__name__}: {result}")
+
     def paint_frametime_graph(self):
         x_offset = self.width()
         height = self.height()
@@ -403,30 +434,42 @@ class Renderer(QFrame):
             current_key = list(objects.keys())[i]
             self.painter.drawText(width - length/2, height - 100 - 6-10*i, "{}: {:.2f}ms".format(current_key, objects[current_key]))
 
-    def draw_line(self, alpha, start, end):
+    def draw_line(self, alpha, start, end, grey_out=False):
         """
         Draws a line at the given alpha level from the start point to the end point.
 
         Arguments:
-            Float alpha: The alpha level from 0.0-1.0 to set the line to.
-                           https://doc.qt.io/qt-5/qcolor.html#alpha-blended-drawing
+            Float alpha: The alpha level (from 0.0 to 1.0) to set the line to.
             List start: The X&Y position of the start of the line.
             List end: The X&Y position of the end of the line.
+            Boolean grey_out: Whether to grey out the line or not.
         """
+        if grey_out:
+            prev_pen = self.painter.pen()
+            PEN_GREY_INACTIVE.setWidth(self.scaled_number(WIDTH_LINE_RAW_VIEW))
+            self.painter.setPen(PEN_GREY_INACTIVE)
+
         self.painter.setOpacity(alpha)
         self.painter.drawLine(self.scaled_point(start[0], start[1]), self.scaled_point(end[0], end[1]))
 
+        if self.raw_view and grey_out:
+            self.painter.setPen(prev_pen)
+
     def draw_cross(self, alpha, point, grey_out, highlight):
         """
         Draws a cross.
 
         Args:
            Float alpha: The alpha level from 0.0-1.0 to set the cross to.
-           List point: The X&Y position of the cross.
+           List point: The X and Y position of the cross.
            Boolean grey_out: Whether to grey out the cross or not.
            Boolean highlight: Whether to highlight the cross or not. This takes
                precedence over ``grey_out`` if both are set.
         """
+        # crosses can clutter the screen sometimes, don't draw them if raw view
+        # is on
+        if self.raw_view:
+            return
         prev_pen = None
         if highlight:
             prev_pen = self.painter.pen()
@@ -458,6 +501,8 @@ class Renderer(QFrame):
             QPainter painter: The painter.
             Hitobj hitobj: A Hitobject.
         """
+        if not self.draw_hitobjects:
+            return
         if isinstance(hitobj, Circle):
             self.draw_hitcircle(hitobj)
             self.draw_approachcircle(hitobj)
@@ -481,11 +526,17 @@ class Renderer(QFrame):
         opacity = max(0, min(1, opacity-fade_out))
         p = hitobj.position
 
+        # the pen width grows outwards and inwards equally (preferring outwards
+        # if the width is odd I think), so we need to tell it to start drawing
+        # half of the pen's width away from the radius for the final circle to
+        # have radius `self.hitcircle_radius`.
+        r = self.scaled_number(self.hitcircle_radius - WIDTH_CIRCLE_BORDER / 2)
+
         PEN_WHITE.setWidth(self.scaled_number(WIDTH_CIRCLE_BORDER))
         self.painter.setOpacity(opacity)
         self.painter.setPen(PEN_WHITE)
         self.painter.setBrush(BRUSH_GRAY)
-        self.painter.drawEllipse(self.scaled_point(p.x, p.y), self.scaled_number(self.hitcircle_radius), self.scaled_number(self.hitcircle_radius))
+        self.painter.drawEllipse(self.scaled_point(p.x, p.y), r, r)
         self.painter.setBrush(BRUSH_BLANK)
 
     @analyzer.track
@@ -521,6 +572,8 @@ class Renderer(QFrame):
             QPainter painter: The painter.
             Hitobj hitobj: A Hitobject.
         """
+        if not self.draw_approach_circles:
+            return
         current_time = self.clock.get_time()
         if self.get_hit_time(hitobj) - current_time < 0:
             return
@@ -528,12 +581,12 @@ class Renderer(QFrame):
         opacity = max(0, min(1, opacity))
         scale = max(1, ((self.get_hit_time(hitobj) - current_time) / self.preempt) * 3 + 1)
         p = hitobj.position
-        radius = self.hitcircle_radius * scale
+        r = self.scaled_number(self.hitcircle_radius * scale)
 
         PEN_WHITE.setWidth(self.scaled_number(WIDTH_CIRCLE_BORDER / 2))
         self.painter.setPen(PEN_WHITE)
         self.painter.setOpacity(opacity)
-        self.painter.drawEllipse(self.scaled_point(p.x, p.y), self.scaled_number(radius), self.scaled_number(radius))
+        self.painter.drawEllipse(self.scaled_point(p.x, p.y), r, r)
 
     @analyzer.track
     def draw_slider(self, hitobj):
@@ -563,7 +616,7 @@ class Renderer(QFrame):
         opacity = max(0, min(1, opacity-fade_out)) * 0.75
         p = hitobj.position
 
-        PEN_GRAY.setWidth(self.scaled_number(self.hitcircle_radius * 2 + WIDTH_CIRCLE_BORDER))
+        PEN_GRAY.setWidth(self.scaled_number(self.hitcircle_radius * 2))
         PEN_GRAY.setCapStyle(Qt.RoundCap)
         PEN_GRAY.setJoinStyle(Qt.RoundJoin)
         self.painter.setPen(PEN_GRAY)
@@ -649,6 +702,12 @@ class Renderer(QFrame):
         if self.paused:
             self.next_frame()
 
+    def wheelEvent(self, event):
+        # support scrolling both vertically or horizontally, just respect
+        # whichever is greatest for that event
+        delta = max(event.pixelDelta().x(), event.pixelDelta().y(), key=abs)
+        self.seek_to(self.clock.time_counter + delta)
+
     def get_hit_endtime(self, hitobj):
         return hitobj.end_time.total_seconds() * 1000 if not isinstance(hitobj, Circle) else self.get_hit_time(hitobj)
 
@@ -671,3 +730,34 @@ class Renderer(QFrame):
 
     def toggle_frametime(self):
         self.paint_frametime = not self.paint_frametime
+
+    def raw_view_changed(self, new_state):
+        self.raw_view = new_state
+        # redraw everything for the new raw view
+        self.update()
+
+    def only_embolden_keydowns_changed(self, new_state):
+        self.only_embolden_keydowns = new_state
+        self.update()
+
+    def hitobjects_changed(self, new_state):
+        self.draw_hitobjects = new_state
+        self.update()
+
+    def approach_circles_changed(self, new_state):
+        self.draw_approach_circles = new_state
+        self.update()
+
+    def num_frames_changed(self, new_value):
+        self.num_frames_on_screen = new_value
+        self.update()
+
+    def circle_size_mod_changed(self, new_value):
+        if not self.has_beatmap:
+            # cs doesn't matter to us if we don't have a beatmap (and we don't
+            # have the attributes necessary to compute it anyway)
+            return
+        use_hr = True if new_value == "HR" else False
+        use_ez = True if new_value == "EZ" else False
+        self.hitcircle_radius = circle_radius(self.beatmap.cs(hard_rock=use_hr, easy=use_ez))
+        self.update()
