@@ -22,8 +22,15 @@ class ReplayInfo(QFrame):
     # in pixels
     EDGE_HIT_THRESH = 3
 
-    def __init__(self, replay, beatmap, slider_dir):
+    def __init__(self, replay, beatmap, slider_dir, ur_result=None, \
+        frametime_result=None, snaps_result=None, hits=None):
+        """
+        If passed, the `ur_result`, `frametime_result`, `snaps_result`, and
+        `hits` parameters will be used instead of recalculating them from
+        scratch.
+        """
         super().__init__()
+        self.replay = replay
         # replay is already loaded so we don't need an api key. We pass a slider
         # dir because `Interface` has already loaded a beatmap for us, but
         # circleguard doesn't know that, so it will redownload the beatmap for
@@ -39,7 +46,7 @@ class ReplayInfo(QFrame):
 
         info_label = QLabel(f"{replay.username} +{mods} on map {replay.map_id}")
 
-        ur_result = circleguard.ur(replay, single=True)
+        ur_result = ur_result or circleguard.ur(replay, single=True)
         ur = round(ur_result.ur, 2)
         ur = self.maybe_highlight(ur, self.UR_YELLOW_THRESH, self.UR_RED_THRESH)
         # highlight ucvUR in the same way as ur or the user will get confused
@@ -49,16 +56,18 @@ class ReplayInfo(QFrame):
 
         ur_label = QLabel(f"<b>cvUR:</b> {ur} ({ucv_ur} ucv)")
 
-        frametime = round(circleguard.frametime(replay, single=True).frametime, 2)
+        frametime_result = frametime_result or circleguard.frametime(replay, single=True)
+        frametime = round(frametime_result.frametime, 2)
         frametime = self.maybe_highlight(frametime, self.FRAMETIME_YELLOW_THRESH, self.FRAMETIME_RED_THRESH)
         frametime_label = QLabel(f"<b>cv frametime:</b> {frametime}")
 
         events_label = QLabel("Events Table")
 
         events = []
-        snaps = circleguard.snaps(replay, single=True).snaps
+        snaps_result = snaps_result or circleguard.snaps(replay, single=True)
+        snaps = snaps_result.snaps
 
-        hits = circleguard.hits(replay)
+        hits = hits or circleguard.hits(replay)
         edge_hits = []
         for hit in hits:
 
@@ -108,6 +117,11 @@ class ReplayInfo(QFrame):
             statistic = f"<font color='red'>{statistic}</font>"
         return statistic
 
+    def __eq__(self, other):
+        if not isinstance(other, ReplayInfo):
+            return False
+        return self.replay == other.replay
+
 class EventsTable(QTableWidget):
     jump_button_clicked = pyqtSignal(int) # time (ms)
 
